from skimage.io import imread
import numpy as np
from utils import eight_bit_stretch

def ndvi(irg, threshold=False):
    """
    Returns the Normalized Difference Vegetation Index for an IRG image
    """
    nir, red = irg[:,:,0].astype(np.float32), irg[:,:,1].astype(np.float32)
    ndvi_with_nan = (nir - red) / (nir + red)
    ndvi = np.nan_to_num(ndvi_with_nan)
    if threshold:
        threshed = np.where(ndvi > 0, ndvi, 0)
        ndvi = eight_bit_stretch(threshed, lower_bound=0, upper_bound=1)
    return ndvi

def ndwi(irg, threshold=0.4):
    """
    Returns the Normalized Difference Water Index for an IRG image
    """
    green, nir = irg[:,:,2].astype(np.float32), irg[:,:,0].astype(np.float32)
    ndwi_with_nan = (green - nir) / (green + nir)
    ndwi_no_nan = np.nan_to_num(ndwi_with_nan)
    if threshold:
        threshed = np.where(ndwi_no_nan > threshold, ndwi_no_nan, threshold)
        ndwi_out = eight_bit_stretch(threshed, lower_bound=threshold, upper_bound=1)
    return ndwi_out

def bare_soil_index(rgb):
    """
    Returns the Bare Soil Index for an RGB image
    """
    r, g, b = rgb[:,:,0].astype(np.float32), rgb[:,:,1].astype(np.float32), rgb[:,:,2].astype(np.float32)
    bi_with_nan = (r + b - g) / (r + b + g)
    bi_no_nan = np.nan_to_num(bi_with_nan)
    stretched = eight_bit_stretch(bi_no_nan, lower_bound=-1, upper_bound=1)
    return stretched


import socket
from warnings import warn

class Tile:

    ROOT_DIR = r'//ordsvy.gov.uk/Data/BCI_Technical/EOResearch'

    PALETTE = {0: (127, 127, 127), # Buildings (blue)
                1: (255, 255, 255), # Cloud (white)
                2: (255, 0, 255), # Other Surface (green)
                3: (255, 0, 0), # Roads (magenta)
                4: (0, 0, 255), # Sea (yellow)     
                5: (0, 255, 0), # Vegetation (cyan)
                6: (0, 0, 255), # Water
                7: (0, 0, 0)} # Other (black)

    def __init__(self, country):
        self.country = country
        self.ROOT_DIR = Tile.ROOT_DIR
        self.palette = Tile.PALETTE
        self.mask_root = self._get_mask_root()

    def _get_mask_root(self):
        if socket.gethostname() == 'ND30187':
            MASK_ROOT = r'O:\LandCover'
        else:
            warn('This might not work on anything other than ND30187')
            MASK_ROOT = r'//afahpoc1w006/Mask_RCNN/LandCover'
        return MASK_ROOT

    def invert_palette(self):
        invert_palette = {v: k for k, v in self.palette.items()}
        return invert_palette

    



import numpy as np

def convert_to_colour(arr_2d, palette):
    """ Numeric labels to RGB-color encoding """
    arr_3d = np.zeros((arr_2d.shape[0], arr_2d.shape[1], 3), dtype=np.uint8)

    for c, i in palette.items():
        m = arr_2d == c
        arr_3d[m] = i

    return arr_3d

def convert_from_colour(arr_3d, palette):
    """ RGB-color encoding to grayscale labels """
    arr_2d = np.zeros((arr_3d.shape[0], arr_3d.shape[1]), dtype=np.uint8)

    for c, i in palette.items():
        m = np.all(arr_3d == np.array(c).reshape(1, 1, 3), axis=2)
        arr_2d[m] = i

    return arr_2d

def eight_bit_stretch(img, lower_bound, upper_bound):
    """ Stretches image to be 8-bit """
    stretched = 255 * (img - lower_bound) / (upper_bound - lower_bound)
    return stretched.astype(np.uint8)




"""
SatZoomer
=========

Provides
    1. Tools to add a new country / satellite specification
    2. Tools to extract the normalized and colour-balanced RGB / IRG tiles from the raw panchromatic mosaic
    3. Tools to combine spectral indices with deep learning heatmaps to produce land cover maps

Available subpackages
---------------------
spectral_indices
    Tools to extract known spectral indices from multispectral data (e.g. NDVI / NDWI)

utils
    Tools to visualize outputs and reduce radiometric depth
"""

__version__ = '1.0.0'


from setuptools import setup
from setuptools import find_packages

long_description = """
Satellite Preprocessing library for BaseMap1
"""

NAME = 'SatZoomer'

requires = ['numpy>=1.9.1',
            'scikit-image',
            'pandas']

classifiers = ['Development Status :: 2 - Pre-Alpha',
                'License :: Other/Proprietary License',
                'Programming Language :: Python :: 3',
                'Operating System :: Microsoft :: Windows',
                'Topic :: Scientific/Engineering :: Visualization']

setup(name=NAME,
        version='1.0.0',
        description='Satellite Image Processing',
        long_description=long_description,
        author='Jacob Rainbow',
        author_email='jacob.rainbow@os.uk',
        url='https://github.com/JRainbowOS/satzoomer',
        install_requires=requires,
        python_requires='>3.6',
        classifiers=classifiers,
        packages=find_packages()
    )

# Needs Python 3.6 due to format strings. This library does not support Python 2.7.


from super_tile import Tile
from spectral_indices import ndvi, ndwi

class GuyanaTile(Tile):

    def __init__(self, tile_number):
        super().__init__('Guyana')
        self.tile_number = tile_number

    def __repr__(self):
        return repr(f'{self.tile_number} in Guyana')

    def _get_imagery_dir(self, bands):
        # os.path.join(self.ROOT_DIR, )
        pass
        

    def _get_mask_dir(self):
 
def main():
    tile_number = 'GT0014'
    tile = GuyanaTile(tile_number)
    print(repr(tile))

if __name__ == '__main__':
    main()



from super_tile import Tile
import os 
from skimage.io import imread, imsave
from spectral_indices import ndvi, ndwi, bare_soil_index

class GhanaTile(Tile):

    STRIDE = 64
    SIZE = 256

    def __init__(self, tile_number):
        super().__init__('Ghana')
        self.tile_number = tile_number
        self.rgb_dir = self._get_imagery_dir('rgb')
        self.irg_dir = self._get_imagery_dir('irg')
        self.mask_dir = self._get_mask_dir()
        self.classes = self._get_classes()
        self.stride, self.size = GhanaTile.STRIDE, GhanaTile.SIZE

    def __repr__(self):
        return repr(f'{self.tile_number} in Ghana')
 
    def _get_imagery_dir(self, bands):
        return os.path.join(self.ROOT_DIR, 'Ghana', 'Techiman', f'{bands.upper()}')        

    def _get_mask_dir(self):
        return os.path.join(self.mask_root, f'guyana_{str(GhanaTile.STRIDE)}_all_layers', f'{self.tile_number}', 'Masks')

    def _get_classes(self):
        os.chdir(self.mask_dir)
        classes = [d.split('_')[0] for d in os.listdir('.') if os.path.isdir(d)]
        return classes 

    def get_image(self, bands, crop_to_size=False):
        if bands == 'irg':
            img_dir = self.irg_dir
        elif bands == 'rgb':
            img_dir = self.rgb_dir
        else:
            raise Exception('Only looking for irg / rgb ')

        img_path = os.path.join(img_dir, f'{self.tile_number}.tif')
        img = imread(img_path)
        if crop_to_size:
            img = img[:9 * self.size, :9 * self.size, :]
        return img

    def get_heatmap(self, class_, crop_to_size=False):
        heatmap_path = os.path.join(self.mask_dir, f'{class_}_masks', f'{self.tile_number}_heatmap.tif')
        heatmap = imread(heatmap_path)
        if crop_to_size:
            padding = self.size - 2 * self.stride
            heatmap = heatmap[padding: padding + 9 * self.size, padding: padding + 9 * self.size]
        return heatmap

    def save_all_dl_heatmaps(self, outdir, crop_to_size=False):
        classes = self.classes
        for class_ in classes:
            hm = self.get_heatmap(class_, crop_to_size=crop_to_size)
            imsave(os.path.join(outdir, f'{self.tile_number}_{class_}.tif'), hm)

    def spectral_index(self, index, threshold=True, crop_to_size=False):       
        if index == 'ndvi':
            img = self.get_image(bands='irg', crop_to_size=crop_to_size)
            out_image = ndvi(img, threshold=threshold)
        elif index == 'ndwi':
            img = self.get_image(bands='irg', crop_to_size=crop_to_size)
            out_image = ndwi(img)
        elif index == 'bi':
            img = self.get_image(bands='rgb', crop_to_size=crop_to_size)
            out_image = bare_soil_index(img)
        return out_image

    def make_landcover_map(self):
        """
        1a. Get Vegetation Heatmap from NDVI
        1b. Take average with Vegetation Heatmap from DL
        2. Get Buildings from DL
        3. Get OtherSurface from DL
        4. Get Roads from DL
        5. Get Water from NDWI (will need fixing)
        6. Take argmax for each pixel ('unclassifiable' if all classes are 0)
        """
        pass


 
def main():
    tile_number = 'TM0032'
    tile = GhanaTile(tile_number)
    print(repr(tile))
    mask_dir = tile.mask_dir
    print(mask_dir)
    classes = tile.classes
    print(classes)
    rgb = tile.get_image(bands='rgb', crop_to_size=True)
    out_dir = r"C:\Users\JRainbow\Documents\Python Scripts\RulesBasedClassification\images"
    # for class_ in classes:
    #     hm = tile.get_heatmap(class_, crop_to_size=True)
    #     imsave(os.path.join(out_dir, f'{tile_number}_{class_}_crop.tif'), hm)    
    # imsave(os.path.join(out_dir, f'{tile_number}_rgb_crop.tif'), rgb)
    # ndvi = tile.spectral_index(index='ndvi', threshold=True, crop_to_size=True)
    ndwi = tile.spectral_index(index='ndwi', threshold=0.5, crop_to_size=True)
    bi = tile.spectral_index(index='bi', crop_to_size=True)
    # imsave(os.path.join(out_dir, f'{tile_number}_ndvi_thresh.tif'), ndvi)
    imsave(os.path.join(out_dir, f'{tile_number}_bi.tif'), bi)



if __name__ == '__main__':
    main()
